@@ -2,7 +2,7 @@ package Locale::CLDR::NumberFormatter;
 
 use version;
 
-our $VERSION = version->declare('v0.26.2');
+our $VERSION = version->declare('v0.27.2');
 
 
 use v5.10;
@@ -15,22 +15,20 @@ use Moose::Role;
 sub format_number {
 	my ($self, $number, $format, $currency, $for_cash) = @_;
 	
-	# Check if we want to format a none numeric number type
-	if(!defined $format) {
-		my $numbering_system = $self->default_numbering_system();
-		if ($self->numbering_system->{$numbering_system}{type} eq 'algorithmic') {
-			$format = $self->numbering_system->{$numbering_system}{data};
-		}
-		else {
-			$format = '0';
-		}
+	# Check if the locales numbering system is algorithmic. If so ignore the format
+	my $numbering_system = $self->default_numbering_system();
+	if ($self->numbering_system->{$numbering_system}{type} eq 'algorithmic') {
+		$format = $self->numbering_system->{$numbering_system}{data};
+		return $self->_algorithmic_number_format($number, $format);
 	}
+	
+	$format //= '0';
+	
+	return $self->_format_number($number, $format, $currency, $for_cash);
+}
 
-	# Some of these algorithmic formats are in locale/type/name format
-	if (my ($locale_id, $type, $format) = $format =~ m(^(.*?)/(.*?)/(.*?)$)) {
-		my $locale = Locale::CLDR->new($locale_id);
-		return $locale->format_number($number, $format);
-	}
+sub _format_number {
+	my ($self, $number, $format, $currency, $for_cash) = @_;
 	
 	# First check to see if this is an algorithmic format
 	my @valid_formats = $self->_get_valid_algorithmic_formats();
@@ -38,7 +36,12 @@ sub format_number {
 	if (grep {$_ eq $format} @valid_formats) {
 		return $self->_algorithmic_number_format($number, $format);
 	}
-		
+	
+	# Some of these algorithmic formats are in locale/type/name format
+	if (my ($locale_id, $type, $format) = $format =~ m(^(.*?)/(.*?)/(.*?)$)) {
+		my $locale = Locale::CLDR->new($locale_id);
+		return $locale->format_number($number, $format);
+	}
 	
 	my $currency_data;
 	
@@ -66,7 +69,8 @@ sub format_number {
 sub add_currency_symbol {
 	my ($self, $format, $symbol) = @_;
 	
-	$format =~ s/¤/$symbol/;
+	
+	$format =~ s/¤/'$symbol'/;
 	
 	return $format;
 }
@@ -108,7 +112,7 @@ sub parse_number_format {
 	$format = $self->add_currency_symbol($format, $currency)
 		if defined $currency;
 	
-	my ($positive, $negative) = $format =~ /^((?:'[^']*')++ | [^';]+)+ (?:;(.+))?$/x;
+	my ($positive, $negative) = $format =~ /^((?:(?:'[^']*')*+[^';]+)+) (?:;(.+))?$/x;
 	
 	my $type = 'positive';
 	foreach my $to_parse ( $positive, $negative ) {
@@ -250,7 +254,16 @@ sub get_formatted_number {
 	
 	my @digits = $self->get_digits;
 	my @number_symbols_bundles = reverse $self->_find_bundle('number_symbols');
-	my %symbols = map { %{$_->number_symbols} } @number_symbols_bundles;
+	my %symbols;
+	foreach my $bundle (@number_symbols_bundles) {
+		my $current_symbols = $bundle->number_symbols;
+		foreach my $type (keys %$current_symbols) {
+			foreach my $symbol (keys %{$current_symbols->{$type}}) {
+				$symbols{$type}{$symbol} = $current_symbols->{$type}{$symbol};
+			}
+		}
+	}
+	
 	my $symbols_type = $self->default_numbering_system;
 	
 	$symbols_type = $symbols{$symbols_type}{alias} if exists $symbols{$symbols_type}{alias};
@@ -277,13 +290,23 @@ sub get_formatted_number {
 	# Handle grouping
 	my ($integer, $decimal) = split /\./, $number;
 
-	my ($separator, $decimal_point) = ($symbols{$symbols_type}{group}, $symbols{$symbols_type}{decimal});
-	my ($minor_group, $major_group) = ($format->{$type}{minor_group}, $format->{$type}{major_group});
+	my $minimum_grouping_digits = $self->_find_bundle('minimum_grouping_digits');
+	$minimum_grouping_digits = $minimum_grouping_digits
+		? $minimum_grouping_digits->minimum_grouping_digits()
+		: 0;
 	
-	if (defined $minor_group) {
-		# Fast commify using unpack
-		my $pattern = "(A$minor_group)(A$major_group)*";
-		$number = reverse join $separator, grep {length} unpack $pattern, reverse $integer;
+	my ($separator, $decimal_point) = ($symbols{$symbols_type}{group}, $symbols{$symbols_type}{decimal});
+	if (($minimum_grouping_digits && length $integer >= $minimum_grouping_digits) || ! $minimum_grouping_digits) {
+		my ($minor_group, $major_group) = ($format->{$type}{minor_group}, $format->{$type}{major_group});
+	
+		if (defined $minor_group && $separator) {
+			# Fast commify using unpack
+			my $pattern = "(A$minor_group)(A$major_group)*";
+			$number = reverse join $separator, grep {length} unpack $pattern, reverse $integer;
+		}
+	}
+	else {
+		$number = $integer;
 	}
 	
 	$number.= "$decimal_point$decimal" if defined $decimal;
@@ -349,6 +372,16 @@ sub _algorithmic_number_format {
 
 sub _get_algorithmic_number_format_data_by_name {
 	my ($self, $format_name, $type) = @_;
+	
+	# Some of these algorithmic formats are in locale/type/name format
+	if (my ($locale_id, undef, $format) = $format_name =~ m(^(.*?)/(.*?)/(.*?)$)) {
+		my $locale = Locale::CLDR->new($locale_id);
+		return $locale->_get_algorithmic_number_format_data_by_name($format, $type)
+			if $locale;
+
+		return undef;
+	}
+	
 	$type //= 'public';
 	
 	my %data = ();
@@ -368,15 +401,35 @@ sub _get_algorithmic_number_format_data_by_name {
 	
 	return keys %data ? \%data : undef;
 }
+
+sub _get_plural_form {
+	my ($self, $plural, $from) = @_;
 	
+	my ($result) = $from =~ /$plural\{(.+?)\}/;
+	($result) = $from =~ /other\{(.+?)\}/ unless defined $result;
+	
+	return $result;
+}
+
 sub _process_algorithmic_number_data {
-	my ($self, $number, $format_data, $in_fraction_rule_set) = @_;
+	my ($self, $number, $format_data, $plural, $in_fraction_rule_set) = @_;
 	
 	$in_fraction_rule_set //= 0;
 	
 	my $format = $self->_get_algorithmic_number_format($number, $format_data);
 	
 	my $format_rule = $format->{rule};
+	if (! $plural && $format_rule =~ /(cardinal|ordinal)/) {
+		my $type = $1;
+		$plural = $self->plural($number, $type);
+		$plural = [$type, $plural];
+	}
+	
+	# Sort out plural forms
+	if ($plural) {
+		$format_rule =~ s/\$\($plural->[0],(.+)\)\$/$self->_get_plural_form($plural->[1],$1)/eg;
+	}
+	
 	my $divisor = $format->{divisor};
 	my $base_value = $format->{base_value} // '';
 	
@@ -386,7 +439,7 @@ sub _process_algorithmic_number_data {
 		$positive_number =~ s/^-//;
 		
 		if ($format_rule =~ /→→/) {
-			$format_rule =~ s/→→/$self->_process_algorithmic_number_data($positive_number, $format_data)/e;
+			$format_rule =~ s/→→/$self->_process_algorithmic_number_data($positive_number, $format_data, $plural)/e;
 		}
 		elsif((my $rule_name) = $format_rule =~ /→(.+)→/) {
 			my $type = 'public';
@@ -396,11 +449,11 @@ sub _process_algorithmic_number_data {
 			my $format_data = $self->_get_algorithmic_number_format_data_by_name($rule_name, $type);
 			if($format_data) {
 				# was a valid name
-				$format_rule =~ s/→(.+)→/$self->_process_algorithmic_number_data($positive_number, $format_data)/e;
+				$format_rule =~ s/→(.+)→/$self->_process_algorithmic_number_data($positive_number, $format_data, $plural)/e;
 			}
 			else {
 				# Assume a format
-				$format_rule =~ s/→(.+)→/$self->format_number($positive_number, $1)/e;
+				$format_rule =~ s/→(.+)→/$self->_format_number($positive_number, $1)/e;
 			}
 		}
 		elsif($format_rule =~ /=%%.*=/) {
@@ -410,13 +463,13 @@ sub _process_algorithmic_number_data {
 			$format_rule =~ s/=%(.*?)=/$self->_algorithmic_number_format($number, $1, 'public')/eg;
 		}
 		elsif($format_rule =~ /=.*=/) {
-			$format_rule =~ s/=(.*?)=/$self->format_number($number, $1)/eg;
+			$format_rule =~ s/=(.*?)=/$self->_format_number($number, $1)/eg;
 		}
 	}
 	# Fractions
 	elsif( $number =~ /\./ ) {
 		my $in_fraction_rule_set = 1;
-		my ($integer, $fraction) = $number =~ /^([^.])*\.(.*)$/;
+		my ($integer, $fraction) = $number =~ /^([^.]*)\.(.*)$/;
 		
 		if ($number >= 0 && $number < 1) {
 			$format_rule =~ s/\[.*\]//;
@@ -426,7 +479,7 @@ sub _process_algorithmic_number_data {
 		}
 		
 		if ($format_rule =~ /→→/) {
-			$format_rule =~ s/→→/$self->_process_algorithmic_number_data_fractions($fraction, $format_data)/e;
+			$format_rule =~ s/→→/$self->_process_algorithmic_number_data_fractions($fraction, $format_data, $plural)/e;
 		}
 		elsif((my $rule_name) = $format_rule =~ /→(.*)→/) {
 			my $type = 'public';
@@ -435,15 +488,15 @@ sub _process_algorithmic_number_data {
 			}
 			my $format_data = $self->_get_algorithmic_number_format_data_by_name($rule_name, $type);
 			if ($format_data) {
-				$format_rule =~ s/→(.*)→/$self->_process_algorithmic_number_data_fractions($fraction, $format_data)/e;
+				$format_rule =~ s/→(.*)→/$self->_process_algorithmic_number_data_fractions($fraction, $format_data, $plural)/e;
 			}
 			else {
-				$format_rule =~ s/→(.*)→/$self->format_number($fraction, $1)/e;
+				$format_rule =~ s/→(.*)→/$self->_format_number($fraction, $1)/e;
 			}
 		}
 		
 		if ($format_rule =~ /←←/) {
-			$format_rule =~ s/←←/$self->_process_algorithmic_number_data($integer, $format_data, $in_fraction_rule_set)/e;
+			$format_rule =~ s/←←/$self->_process_algorithmic_number_data($integer, $format_data, $plural, $in_fraction_rule_set)/e;
 		}
 		elsif((my $rule_name) = $format_rule =~ /←(.+)←/) {
 			my $type = 'public';
@@ -452,10 +505,10 @@ sub _process_algorithmic_number_data {
 			}
 			my $format_data = $self->_get_algorithmic_number_format_data_by_name($rule_name, $type);
 			if ($format_data) {
-				$format_rule =~ s/←(.*)←/$self->_process_algorithmic_number_data($integer, $format_data, $in_fraction_rule_set)/e;
+				$format_rule =~ s/←(.*)←/$self->_process_algorithmic_number_data($integer, $format_data, $plural, $in_fraction_rule_set)/e;
 			}
 			else {
-				$format_rule =~ s/←(.*)←/$self->format_number($integer, $1)/e;
+				$format_rule =~ s/←(.*)←/$self->_format_number($integer, $1)/e;
 			}
 		}
 		
@@ -467,7 +520,7 @@ sub _process_algorithmic_number_data {
 				$format_rule =~ s/=%(.*?)=/$self->_algorithmic_number_format($number, $1, 'public')/eg;
 			}
 			else {
-				$format_rule =~ s/=(.*?)=/$self->format_number($integer, $1)/eg;
+				$format_rule =~ s/=(.*?)=/$self->_format_number($integer, $1)/eg;
 			}
 		}
 	}
@@ -497,18 +550,18 @@ sub _process_algorithmic_number_data {
 					}
 					my $format_data = $self->_get_algorithmic_number_format_data_by_name($rule_name, $type);
 					if ($format_data) {
-						$format_rule =~ s/←(.*)←/$self->_process_algorithmic_number_data($number * $base_value, $format_data, $in_fraction_rule_set)/e;
+						$format_rule =~ s/←(.*)←/$self->_process_algorithmic_number_data($number * $base_value, $format_data, $plural, $in_fraction_rule_set)/e;
 					}
 					else {
-						$format_rule =~ s/←(.*)←/$self->format_number($number * $base_value, $1)/e;
+						$format_rule =~ s/←(.*)←/$self->_format_number($number * $base_value, $1)/e;
 					}
 				}
 				else {
-					$format_rule =~ s/←←/$self->_process_algorithmic_number_data($number * $base_value, $format_data, $in_fraction_rule_set)/e;
+					$format_rule =~ s/←←/$self->_process_algorithmic_number_data($number * $base_value, $format_data, $plural, $in_fraction_rule_set)/e;
 				}
 			}
 			elsif($format_rule =~ /=.*=/) {
-				$format_rule =~ s/=(.*?)=/$self->format_number($number, $1)/eg;
+				$format_rule =~ s/=(.*?)=/$self->_format_number($number, $1)/eg;
 			}
 		}
 		else {
@@ -520,14 +573,14 @@ sub _process_algorithmic_number_data {
 					}
 					my $format_data = $self->_get_algorithmic_number_format_data_by_name($rule_name, $type);
 					if ($format_data) {
-						$format_rule =~ s/→(.+)→/$self->_process_algorithmic_number_data($number % $divisor, $format_data)/e;
+						$format_rule =~ s/→(.+)→/$self->_process_algorithmic_number_data($number % $divisor, $format_data, $plural)/e;
 					}
 					else {
-						$format_rule =~ s/→(.*)→/$self->format_number($number % $divisor, $1)/e;
+						$format_rule =~ s/→(.*)→/$self->_format_number($number % $divisor, $1)/e;
 					}
 				}
 				else {
-					$format_rule =~ s/→→/$self->_process_algorithmic_number_data($number % $divisor, $format_data)/e;
+					$format_rule =~ s/→→/$self->_process_algorithmic_number_data($number % $divisor, $format_data, $plural)/e;
 				}
 			}
 			
@@ -539,14 +592,14 @@ sub _process_algorithmic_number_data {
 					}
 					my $format_data = $self->_get_algorithmic_number_format_data_by_name($rule_name, $type);
 					if ($format_data) {
-						$format_rule =~ s|←(.*)←|$self->_process_algorithmic_number_data(int ($number / $divisor), $format_data)|e;
+						$format_rule =~ s|←(.*)←|$self->_process_algorithmic_number_data(int ($number / $divisor), $format_data, $plural)|e;
 					}
 					else {
-						$format_rule =~ s|←(.*)←|$self->format_number(int($number / $divisor), $1)|e;
+						$format_rule =~ s|←(.*)←|$self->_format_number(int($number / $divisor), $1)|e;
 					}
 				}
 				else {
-					$format_rule =~ s|←←|$self->_process_algorithmic_number_data(int($number / $divisor), $format_data)|e;
+					$format_rule =~ s|←←|$self->_process_algorithmic_number_data(int($number / $divisor), $format_data, $plural)|e;
 				}
 			}
 			
@@ -558,21 +611,21 @@ sub _process_algorithmic_number_data {
 					$format_rule =~ s/=%(.*?)=/$self->_algorithmic_number_format($number, $1, 'public')/eg;
 				}
 				else {
-					$format_rule =~ s/=(.*?)=/$self->format_number($number, $1)/eg;
+					$format_rule =~ s/=(.*?)=/$self->_format_number($number, $1)/eg;
 				}
 			}
 		}
-	}
+	}	
 	
 	return $format_rule;
 }
 
 sub _process_algorithmic_number_data_fractions {
-	my ($self, $fraction, $format_data) = @_;
+	my ($self, $fraction, $format_data, $plural) = @_;
 	
 	my $result = '';
 	foreach my $digit (split //, $fraction) {
-		$result .= $self->_process_algorithmic_number_data($digit, $format_data, 1);
+		$result .= $self->_process_algorithmic_number_data($digit, $format_data, $plural, 1);
 	}
 	
 	return $result;
@@ -601,4 +654,3 @@ no Moose::Role;
 1;
 
 # vim: tabstop=4
-
